/*
  Routeur Solaire pour Linky Mode Standard
  ****************************************
  Version V_3.11_Linky

  Lecture des données du Linky disponibles sur l'interface TIC (Télé-Information Client)
  Linky en mode standard et non mode historique.

  Les détails sont disponibles sur :
  https://f1atb.fr  Section Domotique

  F1ATB Juillet 2023

  GNU General Public License v3.0

  La version  V_3.10_Linky prend en compte le cas des CACSI (pas de valeur injectée) et corrige un bug SINSTS pour le triphasé
  La version  V_3.11_Linky rajoute des sorties MQTT : pws, pwi,triac,action1,action2...
*/

#define HOSTNAME "Linky-ESP32-"
#define CLE_Rom_Init 1234567890  //Valeur pour tester si ROM vierge ou pas. Un changement de valeur remet à zéro toutes les données.


//Librairies
#include <WiFi.h>
#include <ESPmDNS.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include <ArduinoOTA.h>    //Modification On The Air
#include <RemoteDebug.h>   //Debug via Wifi
#include <esp_task_wdt.h>  //Pour un Watchdog
#include <PubSubClient.h>  //Librairie pour la gestion Mqtt
#include "EEPROM.h"        //Librairie pour le stockage en EEPROM historique quotidien

//Program routines
#include "pageHtmlBrute.h"
#include "pageHtmlMain.h"
#include "pageHtmlConnect.h"
#include "pageHtmlPara.h"
#include "pageHtmlActions.h"
#include "Actions.h"


//Watchdog de 120 secondes. Le systeme se Reset si pas de dialoque avec le LINKY ou le Linky distant pendant 120s
#define WDT_TIMEOUT 120

//PINS - GPIO
#define RXD2 26
#define TXD2 27
#define LedYellow 18
#define LedGreen 19
#define pulseTriac 22
#define zeroCross 23


//Nombre Actions Max
#define LesActionsLength 20
//VARIABLES
const char *ap_default_ssid;        // Mode Access point  IP: 192.168.4.1
const char *ap_default_psk = NULL;  // Pas de mot de passe en AP,

//Paramètres dans l'odre de stockage en ROM apres les données du Linky
unsigned long Cle_ROM;
String ssid = "";
String password = "";
byte dhcpOn = 1;
unsigned long IP_Fixe = 0;
unsigned long Gateway = 0;
byte LINKYexterne = 0;
unsigned long LINKYextIP = 0;
unsigned int DomoRepet = 0;
unsigned long DomoIP = 0;
unsigned int DomoPort = 0;
unsigned int DomoIdx = 0;
unsigned int MQTTRepet = 0;
unsigned long MQTTIP = 0;
unsigned int MQTTPort = 0;
String MQTTUser = "User";
String MQTTPwd = "password";



String LinkyExtData;

int adr_debut_para = 0;  //Adresses Para après le Wifi

int WIFIbug = 0;
int IdxDataRawLinky = 0;
int IdxBufDecodLinky = 0;
char DataRawLinky[10000];  //Buffer entrée données Linky


bool LFon = false;
bool EASTvalid = false;
bool EAITvalid = false;
bool DATEvalid = false;


int IdxStockPW = 0;
int SINSTS = 0;
int SINSTI = 0;
long EAST = 0;
long EAIT = 0;
long EAST_last = 0;
long EAIT_last = 0;
long EASJ0 = 0;  //Debut du jour energie active
long EAIJ0 = 0;
float PestimeS = 0;
float PestimeI = 0;
long TlastEASTvalide = 0;
long TlastEAITvalide = 0;
float COSphiS = 1;
float COSphiI = 1;
float PVAcumuleS = 0;
float PVAcumuleI = 0;
int NVAcumulS = 0;
int NVAcumulI = 0;
float moyPWS = 0;
float moyPWI = 0;
float moyPVAS = 0;
float moyPVAI = 0;

int tabPW5mn[600];   //Puissance Active:Soutiré-Injecté toutes les 5mn
float tabPW2s[300];  //Puissance Active:Soutiré-Injecté toutes les 2s
int tabPVA2s[300];   //Puissance Apparente:Soutiré-Injecté toutes les 2s
int IdxStock2s = 0;

Action LesActions[LesActionsLength];  //Liste des actions
int NbActions = 0;


//Internal Timers
unsigned long startMillis;
unsigned long previousWifiMillis;
unsigned long previousHistoryMillis;
unsigned long previousWsMillis;
unsigned long previousWiMillis;
unsigned long previousLinkyExtMillis;
unsigned long previousTimer2sMillis;
unsigned long previousLoop;
unsigned long previousETX;
float previousLoopMin = 1000;
float previousLoopMax = 0;
float previousLoopMoy = 0;
unsigned long previousTimeLinky;
float previousTimeLinkyMin = 1000;
float previousTimeLinkyMax = 0;
float previousTimeLinkyMoy = 0;
unsigned long previousDomoMillis;
unsigned long previousMqttMillis;


String DATE = "";
int HeureCouranteDeci = 0;

//Triac
float retardF = 100;  //Floating value of retard(delay)
//Variables in RAM for interruptions
volatile unsigned long lastIT = 0;
volatile int delayTriac = 0;
volatile int retard = 100;
hw_timer_t *timer = NULL;

WebServer server(80);  // Simple Web Server on port 80

// ************************************

//Debug via WIFI instead of Serial
//Connect a Telnet terminal on port 23
RemoteDebug Debug;
WiFiClient MqttClient;
PubSubClient clientMQTT(MqttClient);

//Multicoeur - Processeur 0 - Collecte données Linky local ou distant
TaskHandle_t Task1;

//Interruptions, Current Zero Crossing from Triac device and Internal Timer
//*************************************************************************
void IRAM_ATTR currentNull() {
  if ((millis() - lastIT) > 2) {  // to avoid glitch detection during 2ms
    delayTriac = 0;               //Time synchro every 10ms
    lastIT = millis();
    digitalWrite(pulseTriac, LOW);  //Stop Triac
  }
}
// Interruption Timer
void IRAM_ATTR onTimer() {  //Interruption every 100 micro second
  delayTriac += 1;
  if (delayTriac > retard && retard < 98) {  //100 steps in 10 ms
    digitalWrite(pulseTriac, HIGH);          //Activate Triac
  }
}

// SETUP
//*******
void setup() {
  startMillis = millis();

  //Pin initialisation
  pinMode(LedYellow, OUTPUT);
  pinMode(LedGreen, OUTPUT);
  pinMode(zeroCross, INPUT);
  pinMode(pulseTriac, OUTPUT);
  digitalWrite(LedYellow, LOW);
  digitalWrite(LedGreen, LOW);
  digitalWrite(pulseTriac, LOW);  //Stop Triac

  //Watchdog initialisation
  esp_task_wdt_init(WDT_TIMEOUT, true);  //enable panic so ESP32 restarts

  //Ports Série
  Serial.begin(115200);
  Serial2.begin(9600, SERIAL_7E1, RXD2, TXD2);  //  7-bit Even parity 1 stop bit pour le Linky
  Serial.println("Booting");

  for (int i = 0; i < LesActionsLength; i++) {
    LesActions[i] = Action(i);  //Creation objets
  }

  esp_task_wdt_reset();

  // Configure WIFI
  // **************
  String hostname(HOSTNAME);
  uint32_t chipId = 0;
  for (int i = 0; i < 17; i = i + 8) {
    chipId |= ((ESP.getEfuseMac() >> (40 - i)) & 0xff) << i;
  }
  hostname += String(chipId);  //Add chip ID to hostname
  WiFi.hostname(hostname);
  Serial.println(hostname);
  ap_default_ssid = (const char *)hostname.c_str();
  // Check WiFi connection
  // ... check mode
  if (WiFi.getMode() != WIFI_STA) {
    WiFi.mode(WIFI_STA);
    delay(10);
  }

  INIT_EEPROM();
  //Lecture Clé pour identifier si la ROM a déjà été initialisée
  Cle_ROM = CLE_Rom_Init;
  unsigned long Rcle = LectureCle();
  Serial.println("cle : " + String(Rcle));
  if (Rcle == Cle_ROM) {  // Programme déjà executé
    LectureEnROM();
    LectureConsoMatinJour();
    InitGpioActions();
  } else {
    RAZ_Histo_Conso();
  }
  Serial.println("SSID:" + ssid);
  Serial.println("Pass:" + password);
  if (ssid.length() > 0) {
    if (dhcpOn == 0) {  //Static IP
      byte arr[4];
      arr[0] = IP_Fixe & 0xFF;          // 0x78
      arr[1] = (IP_Fixe >> 8) & 0xFF;   // 0x56
      arr[2] = (IP_Fixe >> 16) & 0xFF;  // 0x34
      arr[3] = (IP_Fixe >> 24) & 0xFF;  // 0x12
      // Set your Static IP address
      IPAddress local_IP(arr[3], arr[2], arr[1], arr[0]);
      // Set your Gateway IP address
      arr[0] = Gateway & 0xFF;          // 0x78
      arr[1] = (Gateway >> 8) & 0xFF;   // 0x56
      arr[2] = (Gateway >> 16) & 0xFF;  // 0x34
      arr[3] = (Gateway >> 24) & 0xFF;  // 0x12
      IPAddress gateway(arr[3], arr[2], arr[1], arr[0]);
      IPAddress subnet(255, 255, 255, 0);
      IPAddress primaryDNS(arr[3], arr[2], arr[1], arr[0]);  //optional
      IPAddress secondaryDNS(8, 8, 4, 4);                    //optional
      if (!WiFi.config(local_IP, gateway, subnet, primaryDNS, secondaryDNS)) {
        Serial.println("WIFI STA Failed to configure");
      }
    }
    Serial.println("Wifi Begin : " + ssid);
    WiFi.begin(ssid.c_str(), password.c_str());
    while (WiFi.status() != WL_CONNECTED && (millis() - startMillis < 15000)) {  // Attente connexion au Wifi
      Serial.write('.');
      digitalWrite(LedYellow, HIGH);
      digitalWrite(LedGreen, HIGH);
      Serial.print(WiFi.status());
      delay(300);
      digitalWrite(LedYellow, LOW);
      digitalWrite(LedGreen, LOW);
      delay(200);
    }
  }
  if (WiFi.status() == WL_CONNECTED) {
    // ... print IP Address
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
    Serial.println("Connected IP address: " + WiFi.localIP().toString() + " or <a href='http://" + hostname + ".local' >" + hostname + ".local</a>");
  } else {
    Serial.println("Can not connect to WiFi station. Go into AP mode.");
    // Go into software AP mode.
    WiFi.mode(WIFI_AP);
    delay(10);
    WiFi.softAP(ap_default_ssid, ap_default_psk);
    Serial.print("Access Point Mode. IP address: ");
    Serial.println(WiFi.softAPIP());
  }

  // init remote debug
  Debug.begin("ESP32");  // Telnet on port 23
  Debug.println("Ready");
  Debug.print("IP address: ");
  Debug.println(WiFi.localIP());

  //Init Web Server on port 80
  server.on("/", handleRoot);
  server.on("/MainJS", handleMainJS);
  server.on("/Para", handlePara);
  server.on("/ParaJS", handleParaJS);
  server.on("/ParaAjax", handleParaAjax);
  server.on("/ParaUpdate", handleParaUpdate);
  server.on("/Actions", handleActions);
  server.on("/ActionsJS", handleActionsJS);
  server.on("/ActionsUpdate", handleActionsUpdate);
  server.on("/ActionsAjax", handleActionsAjax);
  server.on("/Brute", handleBrute);
  server.on("/BruteJS", handleBruteJS);
  server.on("/ajax_histo48h", handleAjaxHisto48h);
  server.on("/ajax_histo1an", handleAjaxHisto1an);
  server.on("/ajax_dataLinky", handleAjaxLinky);
  server.on("/ajax_dataESP32", handleAjaxESP32);
  server.on("/ajax_data", handleAjaxData);
  server.on("/ajax_data10mn", handleAjaxData10mn);
  server.on("/ajax_etatActions", handleAjax_etatActions);
  server.on("/restart", handleRestart);
  server.onNotFound(handleNotFound);
  server.begin();
  Debug.println("HTTP server started");



  //Interruptions du Triac et Timer interne
  attachInterrupt(zeroCross, currentNull, RISING);

  //Hardware timer
  timer = timerBegin(0, 80, true);  //Clock Divider, 1 micro second Tick
  timerAttachInterrupt(timer, &onTimer, true);
  timerAlarmWrite(timer, 100, true);  //Interrupt every 100 Ticks or microsecond
  timerAlarmEnable(timer);


  // Modification du programme par le Wifi  - OTA(On The Air)
  //***************************************************
  ArduinoOTA.setHostname((const char *)hostname.c_str());
  ArduinoOTA.begin();  //Mandatory

  xTaskCreatePinnedToCore(  //Préparation Tâche Multi Coeur
    Task_LectureLinky,      /* Task function. */
    "Task_LectureLinky",    /* name of task. */
    10000,                  /* Stack size of task */
    NULL,                   /* parameter of the task */
    10,                     /* priority of the task */
    &Task1,                 /* Task handle to keep track of created task */
    0);                     /* pin task to core 0 */

  //Timers
  previousWifiMillis = millis() + 300000;
  previousHistoryMillis = millis() - 290000;
  previousTimer2sMillis = millis();
  previousLoop = millis();
  previousTimeLinky = millis();
  previousDomoMillis = millis();
  previousMqttMillis = millis();
  previousETX = millis();
}

/* **********************
   * ****************** *
   * * Tâches Coeur 0 * *
   * ****************** *
   **********************
*/

void Task_LectureLinky(void *pvParameters) {
  esp_task_wdt_add(NULL);  //add current thread to WDT watch
  esp_task_wdt_reset();
  for (;;) {
    unsigned long tps = millis();
    float deltaT = float(tps - previousTimeLinky);
    previousTimeLinky = tps;
    previousTimeLinkyMin = min(previousTimeLinkyMin, deltaT);
    previousTimeLinkyMin = previousTimeLinkyMin + 0.001;
    previousTimeLinkyMax = max(previousTimeLinkyMax, deltaT);
    previousTimeLinkyMax = previousTimeLinkyMax * 0.9999;
    previousTimeLinkyMoy = deltaT * 0.001 + previousTimeLinkyMoy * 0.999;


    //Recupération des données LINKY
    //******************************
    if (LINKYexterne == 1) {
      if (millis() - previousLinkyExtMillis > 2000) {
        previousLinkyExtMillis = millis();
        CallESP32_Linky();
      }
    } else {
      LectureLinky();
    }
    delay(2);
  }
}

void LectureLinky() {  //Lecture port série du LINKY .
  int V = 0;
  long OldWh = 0;
  float deltaWh = 0;
  float Pmax = 0;
  unsigned long Tm = 0;
  float deltaT = 0;
  while (Serial2.available() > 0) {
    V = Serial2.read();
    DataRawLinky[IdxDataRawLinky] = char(V);
    IdxDataRawLinky = (IdxDataRawLinky + 1) % 10000;
    switch (V) {
      case 2:  //STX (Start Text)
        digitalWrite(LedGreen, LOW);
        digitalWrite(LedYellow, LOW);
        break;
      case 3:  //ETX (End Text)
        previousETX = millis();
        if (SINSTS >= SINSTI) {
          digitalWrite(LedYellow, HIGH);
        } else {
          digitalWrite(LedGreen, HIGH);
        }
        LFon = false;
        break;
      case 10:  // Line Feed. Debut Groupe
        LFon = true;
        IdxBufDecodLinky = IdxDataRawLinky;
        break;
      case 13:       // Line Feed. Debut Groupe
        if (LFon) {  //Debut groupe OK
          LFon = false;
          int nb_tab = 0;
          String code = "";
          String val = "";
          int checksum = 0;
          int checkLinky = -1;
          while (IdxBufDecodLinky != IdxDataRawLinky) {
            if (DataRawLinky[IdxBufDecodLinky] == char(9)) {  //Tabulation
              nb_tab++;
            } else {
              if (nb_tab == 0) {
                code += DataRawLinky[IdxBufDecodLinky];
              }
              if (nb_tab == 1) {
                val += DataRawLinky[IdxBufDecodLinky];
              }
              if (nb_tab <= 1) {
                checksum += (int)DataRawLinky[IdxBufDecodLinky];
              }
            }
            IdxBufDecodLinky = (IdxBufDecodLinky + 1) % 10000;
            if (checkLinky == -1 && nb_tab == 2) {
              checkLinky = (int)DataRawLinky[IdxBufDecodLinky];
              checksum += 18;            //2 tabulations
              checksum = checksum & 63;  //0x3F
              checksum = checksum + 32;  //0x20
            }
          }
          if (code.indexOf("EAST") == 0 || code.indexOf("EAIT") == 0 || code=="SINSTS" || code.indexOf("SINSTI") == 0) {
            if (checksum != checkLinky) {
              Debug.println("Erreur checksum code : " + code + " " + String(checksum) + "," + String(checkLinky));
            }
          }
          if (code.indexOf("EAST") == 0) {
            OldWh = EAST;
            if (OldWh == 0) { OldWh = val.toInt(); }
            EAST = val.toInt();
            Tm = millis();
            deltaT = float(Tm - TlastEASTvalide);
            deltaT = deltaT / float(3600000);
            if (EAST == OldWh) {  //Pas de resultat en Wh
              Pmax = 1.3 / deltaT;
              moyPWS = min(moyPWS, Pmax);
            } else {
              TlastEASTvalide = Tm;
              deltaWh = float(EAST - OldWh);
              moyPWS = 0.1 * deltaWh / deltaT + 0.9 * moyPWS;
            }
            if (EASJ0 == 0) { EASJ0 = EAST; }
            EASTvalid = true;
          }
          if (code.indexOf("EAIT") == 0) {
            OldWh = EAIT;
            if (OldWh == 0) { OldWh = val.toInt(); }
            EAIT = val.toInt();
            Tm = millis();
            deltaT = float(Tm - TlastEAITvalide);
            deltaT = deltaT / float(3600000);
            if (EAIT == OldWh) {  //Pas de resultat en Wh
              Pmax = 1.3 / deltaT;
              moyPWI = min(moyPWI, Pmax);
            } else {
              TlastEAITvalide = Tm;
              deltaWh = float(EAIT - OldWh);
              moyPWI = 0.1 * deltaWh / deltaT + 0.9 * moyPWI;
            }
            if (EAIJ0 == 0) { EAIJ0 = EAIT; }
            EAITvalid = true;
          }
          if (code=="SINSTS") {  //Puissance apparente soutirée. Egalité pour ne pas confondre avec SINSTS1 (triphasé)
            SINSTS = val.toInt();
            moyPVAS = 0.1 * float(SINSTS) + 0.9 * moyPVAS;
            if (moyPVAS > 0) {
              COSphiS = moyPWS / moyPVAS;
              COSphiS = min(float(1.0), COSphiS);
            }
            PestimeS = COSphiS * float(SINSTS);
          }
          if (code.indexOf("SINSTI") == 0) {  //Puissance apparente injectée
            SINSTI = val.toInt();
            moyPVAI = 0.1 * float(SINSTI) + 0.9 * moyPVAI;
            if (moyPVAI > 0) {
              COSphiI = moyPWI / moyPVAI;
              COSphiI = min(float(1.0), COSphiI);
            }
            PestimeI = COSphiI * float(SINSTI);
          }
          if (code.indexOf("DATE") == 0) {
            esp_task_wdt_reset();  //Reset du Watchdog à chaque trame du Linky reçue
            DATE = val;
            DATEvalid = true;
            JourHeureChange();
          }
        }
        break;
      default:
        break;
    }
  }
}
// Client d'un autre ESP32 en charge de decoder le LINKY
//****************
void CallESP32_Linky() {
  int p;
  if (SINSTS >= SINSTI) {
    digitalWrite(LedYellow, HIGH);
  } else {
    digitalWrite(LedGreen, HIGH);
  }
  // Use WiFiClient class to create TCP connections
  WiFiClient clientESP_Linky;
  byte arr[4];
  arr[0] = LINKYextIP & 0xFF;          // 0x78
  arr[1] = (LINKYextIP >> 8) & 0xFF;   // 0x56
  arr[2] = (LINKYextIP >> 16) & 0xFF;  // 0x34
  arr[3] = (LINKYextIP >> 24) & 0xFF;  // 0x12

  String host = String(arr[3]) + "." + String(arr[2]) + "." + String(arr[1]) + "." + String(arr[0]);
  if (!clientESP_Linky.connect(host.c_str(), 80)) {
    //  if (!clientESP_Linky.connect(hostbuf, 80)) {
    Serial.println("connection to clientESP_Linky failed");
    return;
  }
  String url = "/ajax_data";
  clientESP_Linky.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + host + "\r\n" + "Connection: close\r\n\r\n");
  unsigned long timeout = millis();
  while (clientESP_Linky.available() == 0) {
    if (millis() - timeout > 5000) {
      Serial.println(">>> clientESP_Linky Timeout !");
      clientESP_Linky.stop();
      return;
    }
  }

  // Read all the lines of the reply from server
  while (clientESP_Linky.available()) {
    String line = clientESP_Linky.readStringUntil('\r');
    LinkyExtData += line;
    p = LinkyExtData.indexOf("Deb,");
    while (p >= 0) {
      LinkyExtData = LinkyExtData.substring(p + 4);
      p = LinkyExtData.indexOf("Deb,");
    }
    if (LinkyExtData.length() > 30 && LinkyExtData.indexOf("Fin") > 0) {
      int idx = 0;
      String Sval = "";
      while (LinkyExtData.indexOf(",") >= 0) {
        Sval = LinkyExtData.substring(0, LinkyExtData.indexOf(","));
        LinkyExtData = LinkyExtData.substring(LinkyExtData.indexOf(",") + 1);
        switch (idx) {
          case 0:
            DATE = Sval;
            DATEvalid = true;
            JourHeureChange();
            break;
          case 1:
            SINSTS = Sval.toInt();
            break;
          case 2:
            SINSTI = Sval.toInt();
            break;
          case 3:
            PestimeS = Sval.toFloat();
            break;
          case 4:
            PestimeI = Sval.toFloat();
            break;
          case 5:
            EAST = Sval.toInt();
            if (EASJ0 == 0) { EASJ0 = EAST; }
            EASTvalid = true;
            break;
          case 6:
            EAIT = Sval.toInt();
            if (EAIJ0 == 0) { EAIJ0 = EAIT; }
            EAITvalid = true;
            esp_task_wdt_reset();  //Reset du Watchdog à chaque trame reçue
            break;
        }
        idx++;
      }
    }
  }
  if (LinkyExtData.length() > 300) LinkyExtData = "";  //Pas de bonnes données
  digitalWrite(LedGreen, LOW);
  digitalWrite(LedYellow, LOW);
}

/* **********************
   * ****************** *
   * * Tâches Coeur 1 * *
   * ****************** *
   **********************
*/
void loop() {
  //Estimation charge coeur
  long tps = millis();
  float deltaT = float(tps - previousLoop);
  previousLoop = tps;
  previousLoopMin = min(previousLoopMin, deltaT);
  previousLoopMin = previousLoopMin + 0.001;
  previousLoopMax = max(previousLoopMax, deltaT);
  previousLoopMax = previousLoopMax * 0.9999;
  previousLoopMoy = deltaT * 0.001 + previousLoopMoy * 0.999;

  //Gestion des serveurs
  //********************
  ArduinoOTA.handle();
  Debug.handle();
  server.handleClient();

  //Archivage et envois des mesures périodiquement
  //**********************************************
  if (tps - previousHistoryMillis >= 300000) {  //Historique consommation par pas de 5mn
    previousHistoryMillis = tps;
    if (EAST_last == 0) {
      EAST_last = EAST;
    }
    int PWsoutire = 12 * int(EAST - EAST_last);  //Puissance moyenne sur 5mn
    EAST_last = EAST;
    if (EAIT_last == 0) {
      EAIT_last = EAIT;
    }
    int PWinjecte = 12 * int(EAIT - EAIT_last);
    EAIT_last = EAIT;
    int PW = PWsoutire - PWinjecte;  //Puissance signée
    tabPW5mn[IdxStockPW] = PW;
    IdxStockPW = (IdxStockPW + 1) % 600;
  }
  if (tps - previousTimer2sMillis >= 2000) {
    previousTimer2sMillis = tps;
    if (!EAITvalid && tps>8000) { //Cas des CACSI ou EAIT n'est jamais positionné
      EAITvalid=EASTvalid;
    }
    if (EASTvalid && EAITvalid) {
      tabPW2s[IdxStock2s] = PestimeS - PestimeI;
      tabPVA2s[IdxStock2s] = SINSTS - SINSTI;
      //Variation Energie Active
      IdxStock2s = (IdxStock2s + 1) % 300;
      GestionOverproduction();
      int Pw = int(PestimeS - PestimeI);
      EnvoiaDomoticz(Pw);
      envoiAuMQTT(Pw);
    }
  }
  //Vérification du WIFI
  //********************
  if (tps - previousWifiMillis > 30000) {  //Test présence WIFI toutes les 30s
    previousWifiMillis = tps;
    if (WiFi.waitForConnectResult() != WL_CONNECTED) {
      Debug.println("Connection Failed! #" + String(WIFIbug));
      Serial.println("Connection Failed! #" + String(WIFIbug));
      WIFIbug++;
      if (WIFIbug > 20) {
        ESP.restart();
      }
    } else {
      WIFIbug = 0;
    }
    Serial.print("Niveau Signal WIFI:");
    Serial.println(WiFi.RSSI());
    Serial.print("WIFIbug:");
    Serial.println(WIFIbug);
    Debug.print("Niveau Signal WIFI:");
    Debug.println(WiFi.RSSI());
    Debug.print("WIFIbug:");
    Debug.println(WIFIbug);
    Serial.println("Charge Lecture Linky en ms - Min : " + String(int(previousTimeLinkyMin)) + " Moy : " + String(int(previousTimeLinkyMoy)) + "  Max : " + String(int(previousTimeLinkyMax)));
    Debug.println("Charge Lecture Linky en ms - Min : " + String(int(previousTimeLinkyMin)) + " Moy : " + String(int(previousTimeLinkyMoy)) + "  Max : " + String(int(previousTimeLinkyMax)));
    Serial.println("Charge Boucle générale en ms - Min : " + String(int(previousLoopMin)) + " Moy : " + String(int(previousLoopMoy)) + "  Max : " + String(int(previousLoopMax)));
    Debug.println("Charge Boucle générale en ms - Min : " + String(int(previousLoopMin)) + " Moy : " + String(int(previousLoopMoy)) + "  Max : " + String(int(previousLoopMax)));
    int T = int(millis() / 1000);
    float DureeOn = float(T) / 3600;
    Serial.println("ESP32 ON depuis : " + String(DureeOn) + " heures");
    Debug.println("ESP32 ON depuis : " + String(DureeOn) + " heures");
  }
  if ((tps - startMillis) > 180000 && WiFi.getMode() != WIFI_STA) {  //Connecté en  Access Point depuis 3mn. Pas normal
    Serial.println("Pas connecté en WiFi mode Station. Redémarrage");
    delay(5000);
    ESP.restart();
  }
}

// ************
// *  ACTIONS *
// ************
void GestionOverproduction() {
  float Pw = PestimeS - PestimeI;
  //Cas du Triac. Action 0
  if (LesActions[0].Actif == 1 && LesActions[0].TypeEnCours(HeureCouranteDeci) >= 2) {
    float TriacSeuilPw = LesActions[0].Valmin(HeureCouranteDeci);
    float MaxTriac = LesActions[0].Valmax(HeureCouranteDeci);
    float GainBoucle = LesActions[0].Port;                        //Valeur stockée dans Port
    retardF = retardF + 0.1;                                      //On ferme très légèrement si pas de message reçu
    retardF = retardF + (Pw - TriacSeuilPw) * GainBoucle / 1000;  // Gain de boucle de l'asservissement
    if (retardF < 0) { retardF = 0; }
    if (retardF < 100 - MaxTriac) { retardF = 100 - MaxTriac; }
    if (retardF > 100) { retardF = 100; }
    retard = int(retardF);  //Valeure entiere pour piloter le Triac
  } else {
    retardF = 100;
    retard = 100;
  }
  //Cas des GPIOS
  for (int i = 1; i < NbActions; i++) {
    LesActions[i].Activer(Pw, HeureCouranteDeci);
  }
}

void InitGpioActions() {
  for (int i = 1; i < NbActions; i++) {
    if (LesActions[i].Actif == 1) {
      LesActions[i].InitGpio();
    }
  }
}
// ***********
// *  SERVER *
// ***********
void handleRoot() {                  //Pages principales
  if (WiFi.getMode() != WIFI_STA) {  // en AP mode
    String NewSsid = server.arg("ssid");
    NewSsid.trim();
    String NewPassword = server.arg("passe");
    NewPassword.trim();
    Serial.println(NewSsid);
    Serial.println(NewPassword);
    if (NewSsid.length() == 0) {
      server.send(200, "text/html", String(ConnectAP_Html));
    } else {
      String S = String(SwitchWifi_Html);
      S.replace("SSID", NewSsid);
      server.send(200, "text/html", S);
      ssid = NewSsid;
      password = NewPassword;
      EcritureEnROM();
      delay(5000);
      ESP.restart();
    }
  } else {  //Station Mode
    server.send(200, "text/html", String(MainHtml));
  }
}
void handleMainJS() {                             //Code Javascript
  server.send(200, "text/html", String(MainJS));  // Javascript code
}
void handleBrute() {  //Page données brutes
  server.send(200, "text/html", String(PageBrute));
}
void handleBruteJS() {                                 //Code Javascript
  server.send(200, "text/html", String(PageBruteJS));  // Javascript code
}

void handleAjaxLinky() {  // Envoi des dernières données  brutes reçues du LINKY
  String S = "";
  String LinkyExtDataB = "";
  int LastIdx = server.arg(0).toInt();
  if (LINKYexterne == 1) {
    // Use WiFiClient class to create TCP connections
    WiFiClient clientESP_Linky;
    byte arr[4];
    arr[0] = LINKYextIP & 0xFF;          // 0x78
    arr[1] = (LINKYextIP >> 8) & 0xFF;   // 0x56
    arr[2] = (LINKYextIP >> 16) & 0xFF;  // 0x34
    arr[3] = (LINKYextIP >> 24) & 0xFF;  // 0x12

    String host = String(arr[3]) + "." + String(arr[2]) + "." + String(arr[1]) + "." + String(arr[0]);
    if (!clientESP_Linky.connect(host.c_str(), 80)) {
      Serial.println("connection to clientESP_Linky failed");
      return;
    }
    String url = "/ajax_dataLinky?idx=" + String(LastIdx);
    clientESP_Linky.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + host + "\r\n" + "Connection: close\r\n\r\n");
    unsigned long timeout = millis();
    while (clientESP_Linky.available() == 0) {
      if (millis() - timeout > 5000) {
        Serial.println(">>> clientESP_Linky Timeout !");
        clientESP_Linky.stop();
        return;
      }
    }
    // Lecture des données brutes distantes
    while (clientESP_Linky.available()) {
      LinkyExtDataB += clientESP_Linky.readStringUntil('\r');
    }
    S = LinkyExtDataB.substring(LinkyExtDataB.indexOf("\n\n") + 2);
  } else {
    while (LastIdx != IdxDataRawLinky) {
      S += String(DataRawLinky[LastIdx]);
      LastIdx = (1 + LastIdx) % 10000;
    }
    S += "|" + String(IdxDataRawLinky);
  }
  server.send(200, "text/html", S);
}
void handleAjaxHisto48h() {  // Envoi Historique de 50h (600points) toutes les 5mn
  String S = "";
  int iS = IdxStockPW;
  for (int i = 0; i < 600; i++) {
    S += String(tabPW5mn[iS]) + ",";
    iS = (1 + iS) % 600;
  }
  server.send(200, "text/html", S);
}
void handleAjaxESP32() {  // Envoi des dernières infos sur l'ESP32
  String S = "";
  int T = int(millis() / 36000);
  float H = float(T) / 100;
  String coeur0 = String(int(previousTimeLinkyMin)) + ", " + String(int(previousTimeLinkyMoy)) + ", " + String(int(previousTimeLinkyMax));
  String coeur1 = String(int(previousLoopMin)) + ", " + String(int(previousLoopMoy)) + ", " + String(int(previousLoopMax));
  S += String(H) + "|" + WiFi.RSSI() + "|" + WiFi.BSSIDstr() + "|" + ssid + "|" + WiFi.localIP().toString() + "|" + WiFi.gatewayIP().toString() + "|" + WiFi.subnetMask().toString() + "|" + coeur0 + "|" + coeur1;
  server.send(200, "text/html", S);
}
void handleAjaxHisto1an() {  // Envoi Historique Energie quotiiienne sur 1 an 370 points
  server.send(200, "text/html", HistoriqueEnergie1An());
}
void handleAjaxData() {
  long EASJ = EAST - EASJ0;  //Conso du jour
  long EAIJ = EAIT - EAIJ0;
  String S = "Deb," + DATE + "," + String(SINSTS) + "," + String(SINSTI) + "," + String(PestimeS) + "," + String(PestimeI) + "," + String(EAST) + "," + String(EAIT) + "," + String(EASJ) + "," + String(EAIJ) + ",Fin";
  server.send(200, "text/html", S);
}
void handleAjax_etatActions() {
  int NbActifs = 0;
  String S = "";
  String On_;
  for (int i = 0; i < NbActions; i++) {
    if (LesActions[i].Actif == 1) {
      S += String(i) + "," + LesActions[i].Titre + ",";
      if (i == 0) {
        S += String(100 - retard) + ",";
      } else {
        On_ = "Off";
        if (LesActions[i].On) On_ = "On";
        S += On_ + ",";
      }
      S += "|";
      NbActifs++;
    }
  }
  S = String(LINKYexterne) + "|" + String(LINKYextIP) + "|" + NbActifs + "|" + S;
  server.send(200, "text/html", S);
}
void handleRestart() {  // Eventuellement Reseter l'ESP32 à distance
  server.send(200, "text/plain", "OK Reset. Attendez.");
  delay(1000);
  ESP.restart();
}
void handleAjaxData10mn() {  // Envoi Historique de 10mn (300points)Energie Active Soutiré - Injecté
  String S = "";
  int iS = IdxStock2s;
  for (int i = 0; i < 300; i++) {
    S += String(tabPW2s[iS]) + ",";
    S += String(tabPVA2s[iS]) + ",";
    iS = (1 + iS) % 300;
  }
  server.send(200, "text/html", S);
}
void handleActions() {
  server.send(200, "text/html", String(ActionsHtml));
}
void handleActionsJS() {
  server.send(200, "text/html", String(ActionsJS));
}
void handleActionsUpdate() {
  String s = server.arg("actions");
  Serial.println(s);
  String ligne = "";
  NbActions = 0;
  while (s.indexOf("|") >= 0 && NbActions < LesActionsLength) {
    ligne = s.substring(0, s.indexOf("|"));
    s = s.substring(s.indexOf("|") + 1);
    LesActions[NbActions].Definir(ligne);
    NbActions++;
  }
  int adresse_max = EcritureEnROM();
  server.send(200, "text/plain", "OK" + String(adresse_max));
  InitGpioActions();
}
void handleActionsAjax() {
  String S;
  for (int i = 0; i < NbActions; i++) {
    S += LesActions[i].Lire();
  }
  server.send(200, "text/html", S);
}
void handlePara() {
  server.send(200, "text/html", String(ParaHtml));
}
void handleParaUpdate() {
  dhcpOn = byte(server.arg("dhcp").toInt());
  IP_Fixe = strtoul(server.arg("adrIP").c_str(), NULL, 10);
  Gateway = strtoul(server.arg("gateway").c_str(), NULL, 10);
  LINKYexterne = byte(server.arg("LINKYexterne").toInt());
  LINKYextIP = strtoul(server.arg("LINKYextIP").c_str(), NULL, 10);
  DomoRepet = server.arg("DomoRepete").toInt();
  DomoIP = strtoul(server.arg("DomoIP").c_str(), NULL, 10);  //4 bytes
  DomoPort = server.arg("DomoPort").toInt();                 //int 2 bytes
  DomoIdx = server.arg("DomoIdx").toInt();                   // int 2 bytes
  MQTTRepet = server.arg("MQTTRepete").toInt();
  MQTTIP = strtoul(server.arg("MQTTIP").c_str(), NULL, 10);
  MQTTPort = server.arg("MQTTPort").toInt();  //2 bytes
  MQTTUser = DecodeString(server.arg("MQTTUser"));
  MQTTPwd = DecodeString(server.arg("MQTTpwd"));
  int adresse_max = EcritureEnROM();
  server.send(200, "text/plain", "OK" + String(adresse_max));
}
void handleParaJS() {
  server.send(200, "text/html", String(ParaJS));
}
void handleParaAjax() {
  String S = String(dhcpOn) + "," + String(IP_Fixe) + "," + String(Gateway) + "," + String(LINKYexterne) + "," + String(LINKYextIP) + ",";
  S += String(DomoRepet) + "," + String(DomoIP) + "," + String(DomoPort) + "," + String(DomoIdx) + ",";
  S += String(MQTTRepet) + "," + String(MQTTIP) + "," + String(MQTTPort) + "," + EncodeString(MQTTUser) + "," + EncodeString(MQTTPwd);
  server.send(200, "text/html", S);
}
void handleNotFound() {  //Page Web pas trouvé
  Debug.println(F("Fichier non trouvé"));
  String message = "Fichier non trouvé\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}

// *******************
// * DOMOTICZ client * - Option pour transferer vers systeme Domoticz
// *******************
void EnvoiaDomoticz(int PW) {
  unsigned long tps = millis();
  if (int((tps - previousDomoMillis) / 1000) > DomoRepet && DomoRepet != 0) {
    previousDomoMillis = tps;
    byte arr[4];
    arr[0] = DomoIP & 0xFF;          // 0x78
    arr[1] = (DomoIP >> 8) & 0xFF;   // 0x56
    arr[2] = (DomoIP >> 16) & 0xFF;  // 0x34
    arr[3] = (DomoIP >> 24) & 0xFF;  // 0x12

    String host = String(arr[3]) + "." + String(arr[2]) + "." + String(arr[1]) + "." + String(arr[0]);
    WiFiClient DomoClient;
    if (!DomoClient.connect(host.c_str(), DomoPort)) {
      Serial.println("connection vers Domoticz plante");
      return;
    }
    String url;
    url = "/json.htm?type=command&param=udevice&idx=";
    url += String(DomoIdx);
    url += "&nvalue=0&svalue=";
    url += String(PW);
    Serial.println(url);
    // This will send the request to the Domoticz server
    DomoClient.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + host + "\r\n" + "Connection: close\r\n\r\n");
    unsigned long timeout = millis();
    while (DomoClient.available() == 0) {
      if (millis() - timeout > 5000) {
        Serial.println(">>> Client Domoticz Timeout !");
        DomoClient.stop();
        return;
      }
    }
    // Read all the lines of the reply from server and print them to Serial
    while (DomoClient.available()) {
      String line = DomoClient.readStringUntil('\r');
      //No interest of the response
    }
  }
}


// *********
// *  MQTT *
// *********
void envoiAuMQTT(int Pw) {
  int etat = 0;
  unsigned long tps = millis();
  if (int((tps - previousMqttMillis) / 1000) > MQTTRepet && MQTTRepet != 0) {
    previousMqttMillis = tps;
    byte arr[4];
    arr[0] = MQTTIP & 0xFF;          // 0x78
    arr[1] = (MQTTIP >> 8) & 0xFF;   // 0x56
    arr[2] = (MQTTIP >> 16) & 0xFF;  // 0x34
    arr[3] = (MQTTIP >> 24) & 0xFF;  // 0x12

    String host = String(arr[3]) + "." + String(arr[2]) + "." + String(arr[1]) + "." + String(arr[0]);
    clientMQTT.setServer(host.c_str(), MQTTPort);
    //clientMQTT.setServer(hostbuf, MQTTPort);
    clientMQTT.setCallback(callback);  //Déclaration de la fonction de souscription
    if (!clientMQTT.connected()) {
      Serial.println("Connection au serveur MQTT ...");
      if (clientMQTT.connect("LinkyClient", MQTTUser.c_str(), MQTTPwd.c_str())) {
        Serial.println("MQTT connecté");
      } else {
        Serial.print("echec connexion au MQTT, code erreur= ");
        Serial.println(clientMQTT.state());
        return;
      }
    }
    clientMQTT.loop();
    int Pva = SINSTS - SINSTI;
    long EASJ = EAST - EASJ0;  //Conso du jour
    long EAIJ = EAIT - EAIJ0;
    int PWs =0;
    int PWi = 0;
    if (Pw<0) {
      PWi=-Pw; //Injecté
    } else {
      PWs=Pw;
    }
    mqtt_publish("linky/pva", Pva);
    mqtt_publish("linky/easj", EASJ);
    mqtt_publish("linky/eaij", EAIJ);
    mqtt_publish("linky/pw", Pw);
    mqtt_publish("linky/pws", PWs);
    mqtt_publish("linky/pwi", PWi);
    if (LesActions[0].Actif == 1) {  //CAs du Triac
      int triac = 100 - retard;
      mqtt_publish( "linky/triac", triac);
    }
    for (int i = 1; i < NbActions; i++) {
      if (LesActions[i].Actif == 1) {
        etat = 0;
        if (LesActions[i].On) { etat = 1; }
        mqtt_publish( "linky/action" + String(i), etat);  //Etat On Off des actions (relais)
      }
    }
    clientMQTT.loop();
  }
}
//Callback  pour souscrire a un topic et  prévoir une action
void callback(char *topic, byte *payload, unsigned int length) {
  Serial.println("-------Nouveau message du broker mqtt. Non utilisé-----");
}

//Fonction pour publier un float sur un topic
void mqtt_publish(String topic, float t) {
  String t_str = String(t);
  clientMQTT.publish(topic.c_str(), t_str.c_str());
}
// ******
// DIVERS
// ******
String DecodeString(String s) {
  String S = "";
  while (s.length() > 1) {
    S += String((char)s.substring(0, s.indexOf("-")).toInt());
    s = s.substring(s.indexOf("-") + 1);
  }
  return S;
}
String EncodeString(String s) {
  String S = "";
  for (int i = 0; i < s.length(); i++) {
    S += String((int)s.charAt(i)) + "-";
  }
  return S;
}
